from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.invenio.invenio_record_schema import (
    InvenioRecordSchemaBuilder,
)
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack


class RawPreprocessor(PropertyPreprocessor):
    TYPE = "raw"

    @process(
        model_builder=JSONSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "raw",
    )
    def modify_fulltext_schema(self, data, stack: ModelBuilderStack, **kwargs):
        data.pop("type")
        return data

    @process(
        model_builder=MappingBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "raw",
    )
    def modify_fulltext_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "flatten"
        return data

    @process(
        model_builder=InvenioRecordSchemaBuilder,
        path="**/properties/*",
        condition=lambda current, stack: current.type == "raw",
    )
    def modify_fulltext_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        data["type"] = "raw"
        return data



try:
    import json5
except ImportError:
    import json as json5


def json_loader(file_path, schema, content=None):
    if content:
        return json5.loads(content)
    with open(file_path) as f:
        return json5.load(f)


def yaml_loader(file_path, schema, content=None):
    try:
        import yaml
    except ImportError:
        raise Exception(
            "Loader for yaml not found. Please install pyyaml library to use yaml files"
        )

    if content:
        return yaml.full_load(content)

    with open(file_path) as f:
        return yaml.full_load(f)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_config"
    class_config = "config-dummy-class"
    template = "config"



from functools import cached_property

from deepdiff import DeepDiff

from .schema import Ref, SchemaPathValidator, model_paths


class ReplaceElement(Exception):
    def __init__(self, data):
        super().__init__()
        self.data = data


class RemoveElement(ReplaceElement):
    def __init__(self):
        super().__init__(None)


class ModelBuilderStackEntry:
    key: str
    schema: SchemaPathValidator

    def __init__(self, key: str, data, schema: SchemaPathValidator):
        self.key = key
        self.data = data
        self.schema = schema

    def __getitem__(self, item):
        return self.data[item]

    def __eq__(self, other):
        return self.key == other.key and not DeepDiff(self.key, other.key)

    def __str__(self):
        return f"{self.key} - {self.data}"

    @property
    def schema_valid(self):
        return self.schema.valid

    @property
    def schema_element_type(self):
        if self.schema_valid and isinstance(self.schema, Ref):
            return self.schema.element_type
        return None


class ModelBuilderStack:
    DICT = "dict"
    LIST = "list"
    PRIMITIVE = "primitive"
    SKIP = "skip"

    def __init__(self):
        self.stack = []

    def __getitem__(self, item):
        return self.stack[item]

    def push(self, key, el):
        self._clear_path()
        if not self.stack:
            entry = ModelBuilderStackEntry(key, el, model_paths)
        else:
            try:
                entry = ModelBuilderStackEntry(key, el, self.top.schema.get(key))
            except Exception as e:
                print(self.top.schema.get(key))
        self.stack.append(entry)

    def pop(self):
        self._clear_path()
        self.stack.pop()

    @property
    def top(self):
        return self.stack[-1]

    @property
    def level(self):
        return len(self.stack)

    @property
    def top_type(self):
        match self.top.data:
            case dict():
                return self.DICT
            case list():
                return self.LIST
            case _:
                return self.PRIMITIVE

    @cached_property
    def path(self):
        return "/" + "/".join(str(x.key) for x in self.stack if x.key)

    def _clear_path(self):
        if "path" in self.__dict__:
            del self.__dict__["path"]

    @property
    def schema_valid(self):
        if not self.stack:
            return False
        return self.top.schema_valid



TEMPLATES = {
    "record": "templates/invenio_record.py.jinja2",
    "record-metadata": "templates/invenio_record_metadata.py.jinja2",
    "record-permissions": "templates/invenio_record_permissions.py.jinja2",
    "record-search-options": "templates/invenio_record_search_options.py.jinja2",
    "record-service-config": "templates/invenio_record_service_config.py.jinja2",
    "record-service": "templates/invenio_record_service.py.jinja2",
    "record-schema": "templates/invenio_record_schema.py.jinja2",
    "object-schema": "templates/invenio_record_object_schema.py.jinja2",
    "record-dumper": "templates/invenio_record_dumper.py.jinja2",
    "record-resource-config": "templates/invenio_record_resource_config.py.jinja2",
    "record-resource": "templates/invenio_record_resource.py.jinja2",
    "views": "templates/invenio_views.py.jinja2",
    "invenio_version": "templates/invenio_version.py.jinja2",
    "config": "templates/invenio_config.py.jinja2",
    "ext": "templates/invenio_ext.py.jinja2",
    "proxies": "templates/invenio_proxies.py.jinja2",
    "record-facets": "templates/invenio_record_facets.py.jinja2",
    # utils and included
    "imports": "templates/imports.py.jinja2",
    # scripts
    "script-import-sample-data": "templates/script_import_sample_data.py.jinja2",
    "script-import-sample-data-shell": "templates/script_import_sample_data.sh.jinja2",
}



import re


def camel_case(s):
    s = re.split("_|-", s)
    s = [x[0].title() + x[1:] for x in s]
    return "".join(s)


def snake_case(s):
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    s = re.sub("__([A-Z])", r"_\1", s)
    s = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s)
    return s.lower()



class ModelPreprocessor:
    def __init__(self, builder: "oarepo_model_builder.builder.ModelBuilder"):
        self.builder = builder

    def transform(self, schema, settings):
        pass

    def set(self, settings, name, callable=None):
        if callable:
            if not settings.get(name):
                settings[name] = callable()
            return

        def w(func):
            if not settings.get(name):
                settings[name] = func()

        return w



from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioExtSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_ext_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        ext_class = self.settings.python.ext_class.rsplit(".", maxsplit=1)

        output.add_entry_point(
            "invenio_base.api_apps",
            self.settings.package,
            f"{ext_class[0]}:{ext_class[-1]}",
        )



from oarepo_model_builder.builders.inherited_model import InheritedModelBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack, ReplaceElement
from oarepo_model_builder.stack.stack import RemoveElement


class InheritedModelPreprocessor(PropertyPreprocessor):
    TYPE = "inherited_model_preprocessor"

    @process(model_builder=InheritedModelBuilder, path="/settings$")
    def replace_settings(self, data, stack: ModelBuilderStack, **kwargs):
        if "for-inheritance" in data:
            return data
        new_settings = self._replace_settings_for_inheritance(data)
        new_settings["for-inheritance"] = True
        raise ReplaceElement({"settings": new_settings})

    def _replace_settings_for_inheritance(self, settings):
        new_settings = {}
        for k, v in settings.items():
            if k.endswith("-class"):
                new_settings[k.replace("-class", "-base-classes")] = [v]
            if isinstance(v, dict):
                new_settings[k] = self._replace_settings_for_inheritance(v)
        return new_settings

    @process(model_builder=InheritedModelBuilder, path="/output-directory$")
    def remove_output_directory(self, data, stack: ModelBuilderStack, **kwargs):
        raise RemoveElement()



from oarepo_model_builder.builders import process
from oarepo_model_builder.utils.jinja import package_name

from ..outputs.json_stack import JSONStack
from ..utils.deepmerge import deepmerge
from ..utils.hyphen_munch import HyphenMunch
from .invenio_base import InvenioBaseClassPythonBuilder

OAREPO_FACETS_PROPERTY = "oarepo:facets"
OAREPO_SORTABLE_PROPERTY = "oarepo:sortable"


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search"
    class_config = "record-search-options-class"
    template = None

    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.template = "record-search-options"
        self.search_options_data = []
        self.sort_options_data = []
        self.search_facets_definiton = []
        self.search_options_stack = JSONStack()
        self.facets_definition = []
        self.facets_names = []
        self.settings = settings
        if "oarepo:sortable" in schema:
            self.process_top_sortable(schema["oarepo:sortable"])

    def finish(self, **extra_kwargs):
        super().finish(
            search_options_data=self.search_options_data,
            facets_definition=self.facets_definition,
            sort_definition=self.sort_options_data,
        )
        python_path = self.class_to_path(self.settings.python["record-facets-class"])
        self.process_template(
            python_path,
            "record-facets",
            current_package_name=package_name(
                self.settings.python["record-facets-class"]
            ),
            search_options_data=self.search_options_data,
            **extra_kwargs,
        )

    def process_top_sortable(self, data):
        keys = dir(data)
        for k in keys:
            fields = dir(data[k])
            fields_options = {}
            for field in fields:
                fields_options = deepmerge(fields_options, {field: data[k][field]})
            self.sort_options_data.append({k: fields_options})

    @process("/model/**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        schema_element_type = self.stack.top.schema_element_type

        definition = None
        recurse = True

        if recurse:
            # process children
            self.build_children()

        data = self.stack.top.data

        if not self.search_options_stack:
            return

        if schema_element_type == "property":
            sort_definition = data.get(OAREPO_SORTABLE_PROPERTY, None)

            if sort_definition != None:
                self.sort_options_data.append(
                    self.process_sort_options(self.stack.path, sort_definition)
                )
        array_items_type = None
        if schema_element_type == "property" and data.type == "array":
            try:
                array_items_type = data["items"]["type"]
            except:
                array_items_type = None

        if (
            schema_element_type == "property"
            and data.type != "text"
            and data.type != "fulltext"
            and data.type != "object"
            and data.type != "nested"
            and not (data.type == "array" and array_items_type == "fulltext")
        ):
            definition = data.get(OAREPO_FACETS_PROPERTY, {})
            nested_paths = []
            nested_path = ""
            nested = False
            path_stack = self.stack.stack[3:]  # start inside model properties
            for upper in path_stack:
                if upper.key == "properties":
                    continue
                nested_path = nested_path + upper.key + "."
                if (
                    upper.data.get("oarepo:mapping", HyphenMunch({"type": ""})).type
                    == "nested"
                ):
                    nested_paths.append(nested_path)
            if len(nested_paths) > 0:
                nested = True

            if "key" in definition:
                name = definition["key"]
            else:
                name = self.process_name(self.stack.path, type="name")
            if data.type == "fulltext+keyword" and "key" not in definition:
                name = name + "_keyword"
            if name == "$schema":
                name = "_schema"
            if name == "id":
                name = "_id"
            class_string = ""

            if nested:
                class_string = "NestedLabeledFacet("
                for path in nested_paths:
                    if nested_paths[-1] == path:
                        class_string = class_string + "path = " + '"' + path[:-1] + '"'
                    else:
                        class_string = (
                            class_string
                            + "path = "
                            + '"'
                            + path[:-1]
                            + '"'
                            + ", nested_facet = NestedLabeledFacet("
                        )

            if "field" in definition:
                field = definition["field"]
                if nested:
                    class_string = class_string + " , nested_facet =" + field
                    for x in nested_paths:
                        class_string = class_string + ")"
                    self.search_options_data.append({name: class_string})
                else:
                    self.search_options_data.append({name: field})
            else:
                search_data = []
                field = self.process_name(self.stack.path, type="field")
                if data.type == "fulltext+keyword":
                    field = field + ".keyword"
                search_data.append(["field", field])
                facets_class = definition.get("class", "TermsFacet")
                for key, value in definition.items():
                    if "class" != key and "field" != key:
                        search_data.append([key, value])
                if nested:
                    search_options = self.process_search_options(
                        search_data, facets_class
                    )
                    search_options = class_string + " , nested_facet =" + search_options
                    for x in nested_paths:
                        search_options = search_options + ")"

                else:
                    search_options = self.process_search_options(
                        search_data, facets_class
                    )
                self.search_options_data.append({name: search_options})
            facets_name = "facets." + name
            self.facets_definition.append({name: facets_name})

    def process_search_options(self, data, field_class):
        text = ""
        for x in data:
            if text == "":
                text = text + x[0] + ' = "' + x[1] + '"'
            else:
                text = text + ", " + x[0] + ' = "' + x[1] + '"'
        return field_class + "(" + text + ")"

    def process_name(self, path, type):
        path_array = (path.split("/"))[3:]
        name = path_array[0]
        if len(path_array) == 1:
            return name
        path_array.pop(0)
        last_path = ""
        for path in path_array:
            if last_path != "properties" and path == "items":
                continue
            last_path = path
            if path == "properties":
                continue
            if type == "name":
                name = name + "_" + path
            elif type == "field":
                name = name + "." + path

        return name

    def process_sort_options(self, path, definition):
        field = self.process_name(path=path, type="field")

        key = definition.get("key", "")
        if key == "":
            key = self.process_name(path, "name")
        order = definition.get("order", "asc")
        if order == "desc":
            field = "-" + field

        return {key: dict(fields=[field])}



from oarepo_model_builder.builders import OutputBuilder, process
from oarepo_model_builder.stack import ModelBuilderStack


class JSONBaseBuilder(OutputBuilder):
    output_file_name: str = None
    output_file_type: str = None
    parent_module_root_name: str = None

    def model_element_enter(self):
        top = self.stack.top
        data = top.data
        data = self.call_components("model_element_enter", data, stack=self.stack)
        match self.stack.top_type:
            case self.stack.PRIMITIVE:
                self.output_primitive(top, data)
            case self.stack.LIST:
                self.output.enter(top.key, [])
            case self.stack.DICT:
                self.output.enter(top.key, {})

    def output_primitive(self, top, data):
        self.output.primitive(top.key, data)

    def model_element_leave(self):
        self.call_components(
            "model_element_leave", self.stack.top.data, stack=self.stack
        )
        if self.stack.top_type != self.stack.PRIMITIVE:
            self.output.leave()

    @process("/model")
    def enter_model(self):
        output_name = self.settings[self.output_file_name]
        self.output = self.builder.get_output(self.output_file_type, output_name)
        self.on_enter_model(output_name)
        self.build_children()
        self.on_leave_model()

    def on_enter_model(self, output_name):
        pass

    def on_leave_model(self):
        pass



from oarepo_model_builder.builders import process
from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.builders.utils import ensure_parent_modules
from oarepo_model_builder.stack import ModelBuilderStack


class PythonStructureBuilder(PythonBuilder):
    TYPE = "python_structure"

    @process("/model")
    def model(self):
        self.build_children()
        package_path = self.settings.package_path

        ensure_parent_modules(
            self.builder,
            self.builder.output_dir.joinpath(package_path / "__init__.py"),
            max_depth=len(package_path.parts),
        )



import json

import yaml
from deepdiff import DeepDiff

from ..utils.verbose import log
from . import OutputBase
from .json_stack import JSONStack


class YAMLOutput(OutputBase):
    IGNORE_NODE = JSONStack.IGNORED_NODE
    IGNORE_SUBTREE = JSONStack.IGNORED_SUBTREE
    TYPE = "yaml"

    def begin(self):
        self._created = True
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = yaml.load_all(f, yaml.SafeLoader)  # noqa
                self._created = False
        except FileNotFoundError:
            self.original_data = None
        except ValueError:
            self.original_data = None

        self.stack = JSONStack()
        self.documents = []

    def next_document(self):
        if self.stack.value:
            self.documents.append(self.stack.value)
            self.stack = JSONStack()

    def finish(self):
        self.next_document()

        if not self.documents and self.original_data:
            # nothing generated, keep the file
            return

        self._created = False
        if DeepDiff(self.documents, self.original_data):
            self.builder.filesystem.mkdir(self.path.parent)
            log(2, "Saving %s", self.path)
            with self.builder.filesystem.open(self.path, mode="w") as f:
                yaml.safe_dump_all(self.documents, f, allow_unicode=True)

    @property
    def created(self):
        return self._created

    @created.setter
    def created(self, value):
        self._created = value

    def enter(self, key, el):
        if key is not None:
            self.stack.push(key, el)

    def leave(self):
        if not self.stack.empty:
            self.stack.pop()

    def primitive(self, key, value):
        if key is not None:
            self.stack.push(key, value)
            self.stack.pop()

    def merge(self, value):
        self.stack.merge(value)



from .invenio_base import InvenioBaseClassPythonBuilder
from ..builders import process


class InvenioRecordBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record"
    class_config = "record-class"
    template = "record"

    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.relations = []

    @process("/model/**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        self.build_children()
        data = self.stack.top.data
        if isinstance(data, dict) and "invenio:relation" in data:
            self.relations.append(data["invenio:relation"])

    def process_template(self, python_path, template, **extra_kwargs):
        return super().process_template(
            python_path,
            template,
            **{**extra_kwargs, "invenio_relations": self.relations}
        )
